use std::{
    collections::{BTreeSet, HashMap},
    fs,
    io::ErrorKind,
    path::{Path, PathBuf},
};

use lazy_static::lazy_static;
use log::{debug, error, trace, warn};
use regex::Regex;

use crate::config::Shows;

lazy_static! {
    static ref NUMBER_REGEX: Regex = Regex::new("[0-9]+").unwrap();
}

/// [episode number, or None for a standalone movie] -> [video file path]
type Enumeration = HashMap<Option<u32>, PathBuf>;

/// [show name] -> [enumeration]
type EnumeratedShows = HashMap<String, Enumeration>;

pub fn enumerate_shows(shows: Shows) -> EnumeratedShows {
    shows
        .into_iter()
        .filter_map(|(name, path)| {
            debug!("Enumerating {} from {}", name, path.display());
            enumerate_show(path)
                .map_err(|e| {
                    error!("Error processing {}: {}", name, e);
                })
                .ok()
                .map(|en| (name, en))
        })
        .collect()
}

fn enumerate_show(path: PathBuf) -> std::io::Result<Enumeration> {
    let metadata = fs::metadata(&path)?;
    if metadata.is_file() {
        debug!("{} is a file, standalone", path.display());
        Ok(HashMap::from([(None, path)]))
    } else if metadata.is_dir() {
        debug!("{} is a directory, enumerating episodes", path.display());
        enumerate_dir(path)
    } else {
        Err(std::io::Error::new(
            ErrorKind::InvalidInput,
            format!("Invalid file type for {}", path.display()),
        ))
    }
}

fn enumerate_dir(path: PathBuf) -> std::io::Result<Enumeration> {
    let files: Vec<(PathBuf, String)> = fs::read_dir(&path)?
        .map(|entry| {
            let entry = entry?;
            if !entry.file_type()?.is_file() {
                debug!("Skipping {}, not a file", entry.path().display());
                return Ok(None);
            }
            Ok(Some((
                entry.path(),
                entry.file_name().into_string().map_err(|o| {
                    std::io::Error::new(ErrorKind::InvalidData, format!("Invalid filename {:?}", o))
                })?,
            )))
        })
        .filter_map(|r| r.transpose())
        .collect::<std::io::Result<Vec<(PathBuf, String)>>>()?;

    // For each string prefix which is followed by a number in one or
    // more filenames, record what the number is for each filename
    // with that prefix, sorted by number.
    //
    // We use a BTreeSet of pairs rather than a BTreeMap so that
    // duplicates are preserved.
    let mut prefix_enumerations: HashMap<&str, BTreeSet<(u32, &Path)>> = HashMap::new();
    for (path, filename) in files.iter() {
        for m in NUMBER_REGEX.find_iter(filename) {
            let num = m.as_str().parse();
            if let Err(e) = num {
                warn!("Failed to parse candidate number: {}", e);
                continue;
            }
            let num = num.unwrap();
            let prefix = filename.get(..m.start()).unwrap();
            trace!("{}: candidate prefix {}, number {}", filename, prefix, num);
            prefix_enumerations
                .entry(prefix)
                .or_default()
                .insert((num, path));
        }
    }

    // Use the prefix enumeration that produces the most consecutive numbers
    let best_enumeration = prefix_enumerations
        .into_values()
        .max_by_key(|set| {
            set.iter()
                .zip(set.iter().skip(1))
                .filter(|((cur, _), (next, _))| next - cur == 1)
                .count()
        })
        .ok_or(std::io::Error::new(
            ErrorKind::InvalidData,
            "No valid prefixes found",
        ))?;

    let mut result: Enumeration = HashMap::new();
    for (num, path) in best_enumeration.into_iter() {
        debug!("Episode {}: {}", num, path.display());
        if let Some(dup) = result.insert(Some(num), path.to_path_buf()) {
            warn!(
                "Duplicate episode number, discarding file {}",
                dup.display()
            );
        }
    }
    Ok(result)
}
