use std::path::Path;

use enumeration::enumerate_shows;
use log::{debug, error, info};
use rand::{distributions::Standard, seq::IteratorRandom, Rng};

mod config;
mod enumeration;
mod video;

#[tokio::main]
async fn main() -> std::io::Result<()> {
    env_logger::init();
    ffmpeg_next::init().unwrap();
    dotenvy::dotenv().ok();

    let mut rng = rand::thread_rng();

    let conf = config::load();
    let enumerated_shows = enumerate_shows(conf.shows);

    loop {
        let (show, episodes) = enumerated_shows
            .iter()
            .choose(&mut rng)
            .expect("No shows found!");
        let (num, file) = episodes.iter().choose(&mut rng).unwrap();

        let descriptor = format!(
            "{}{}",
            show,
            if let Some(n) = num {
                format!(" episode {}", n)
            } else {
                String::new()
            }
        );

        info!("Selected: {} - {}", descriptor, file.display());

        let video_info = video::get_video_info(file, Some("eng")).unwrap();
        debug!(
            "Video duration: {}",
            format_timestamp(video_info.duration_secs, None)
        );
        debug!(
            "Subtitle stream index: {:?}",
            video_info.subtitle_stream_index
        );

        let timestamp = video_info.duration_secs * rng.sample::<f64, _>(Standard);
        let formatted_timestamp = format_timestamp(timestamp, Some(video_info.duration_secs));
        info!("Taking screencap at {}", formatted_timestamp);
        if let Err(e) = video::take_screencap(
            file,
            timestamp,
            video_info.subtitle_stream_index,
            &Path::new("./out.png"),
        ) {
            error!("Failed to take screencap: {}", e);
        }

        break;
    }
    Ok(())
}

fn format_timestamp(timestamp: f64, total_duration: Option<f64>) -> String {
    let total_duration = total_duration.unwrap_or(timestamp);
    format!(
        "{}{:02}:{:05.2}",
        if total_duration >= 3600.0 {
            format!("{}:", (timestamp / 3600.0) as u32)
        } else {
            String::new()
        },
        ((timestamp % 3600.0) / 60.0) as u32,
        timestamp % 60.0
    )
}
